pub mod config;
pub mod server;
pub mod client;
pub mod protocol;

use std::{net::ToSocketAddrs, sync::{atomic::AtomicBool, Arc}};
use smol::spawn;
use crate::config::Config;

pub const BUF: usize = 128;

fn main() {
    let args = std::env::args().skip(1).collect::<Vec<_>>();
    if args.len() != 1 && args.len() != 3 {
        eprintln!("---- 🦋 tmoth 🦋 ----");
        eprintln!("tmoth keeps tcp connections alive,");
        eprintln!("even when they are interrupted.");
        eprintln!("it even supports reconnecting.");
        eprintln!("    --  Usage  --");
        eprintln!("tmoth <config>    # server mode");
        eprintln!("tmoth <local:port> <remote:port> <name>");
        return;
    }
    if args.len() == 1 {
        let config = std::fs::read_to_string(&args[0]).expect("failed to read config");
        let config = Config::parse(&config).expect("failed to parse config");
        smol::block_on(server::run(config));
    } else {
        smol::block_on(async {
            let kill = Some(Arc::new(AtomicBool::new(false)));
            let mut tasks = vec![];
            for addr_from in args[0].to_socket_addrs().unwrap() {
                eprintln!("{}: {}", args[0], addr_from);
                let addrs = args[1].to_socket_addrs().unwrap().collect::<Vec<_>>();
                let addr = if addrs.len() == 1 {
                    addrs[0]
                } else if addrs.is_empty() {
                    panic!("{}: no addresses", args[1]);
                } else {
                    for (i, addr) in addrs.iter().enumerate() {
                        eprintln!("> {}: {addr}", i+1);
                    }
                    loop {
                        eprintln!("Enter a number (1-{})", addrs.len());
                        if let Some(line) = std::io::stdin().lines().next().and_then(|v| v.ok()) {
                            let line = line.trim();
                            let index = if line.is_empty() {
                                1
                            } else {
                                line.parse::<usize>().expect("Expected a number")
                            };
                            if index != 0 && index <= addrs.len() {
                                break addrs[index-1];
                            }
                        } else {
                            break addrs[0];
                        }
                    }
                };
                tasks.push(spawn(client::run(addr_from, addr, args[2].clone(), kill.as_ref().map(Arc::clone))));
            }
            let mut line = String::new();
            while async_std::io::stdin().read_line(&mut line).await.is_ok() {
                let line = std::mem::replace(&mut line, String::new());
                let line = line.trim();
                match line {
                    "kill" => {
                        if let Some(kill) = &kill {
                            eprintln!("[`kill`] killing connections on next forwarding action. `unkill` will revert this.");
                            kill.store(true, std::sync::atomic::Ordering::Relaxed);
                        } else {
                            eprintln!("[`kill`] cannot kill");
                        }
                    }
                    "unkill" => {
                        if let Some(kill) = &kill {
                            eprintln!("[`unkill`] connections will no longer be killed.");
                            kill.store(false, std::sync::atomic::Ordering::Relaxed);
                        } else {
                            eprintln!("[`unkill`] cannot kill/unkill");
                        }
                    }
                    _ => {
                        eprintln!("[`{line}`] Unknown command, try `kill` / `unkill`");
                    }
                }
            }
            eprintln!("stdin closed, running noninteractively");
            for task in tasks {
                task.await;
            }
        });
    }
}
