use std::collections::{HashMap, VecDeque};
use std::sync::atomic::{AtomicBool, Ordering};
use std::{net::SocketAddr, time::Duration};
use std::sync::Arc;
use smol::channel::{unbounded, Receiver, Sender};
use smol::lock::Mutex;
use smol::{spawn, Task};
use smol::{io::{self, AsyncReadExt, AsyncWriteExt}, net::TcpStream, Timer};

use crate::BUF;

pub const MAJOR: u8 = 0;
pub const MINOR: u8 = 0;
pub const PATCH: u8 = 0;

pub async fn server(mut connection: TcpStream, mut connection_id: u64, mapping: Arc<Vec<(String, String)>>, connections: Arc<Mutex<HashMap<u64, (Task<()>, Arc<Mutex<(VecDeque<u8>, Sender<Result<bool, VecDeque<u8>>>, Receiver<Result<bool, VecDeque<u8>>>, TcpStream)>>)>>>) -> io::Result<()> {
    connection.write_all(&[b't', b'm', b'o', b't', b'h', MAJOR, MINOR, PATCH]).await?;
    connection.write_all(&connection_id.to_be_bytes()).await?;
    let mut b = [0u8];
    connection.read_exact(&mut b).await?;
    let mutex = if b[0] < 255 {
        // new connection
        let mut name_buf = vec![0u8; b[0] as usize];
        connection.read_exact(&mut name_buf).await?;
        let name = String::from_utf8(name_buf).map_err(|_| io::ErrorKind::InvalidData)?;
        let mut addr = None;
        for (mname, maddr) in mapping.iter() {
            if mname == name.as_str() {
                addr = Some(maddr.to_owned());
                break;
            }
        }
        let o = if let Some(addr) = addr {
            match TcpStream::connect(&addr).await {
                Ok(mut target) => {
                    // forward between connection and target, with buffers in case of disconnect
                    let (send, recv) = unbounded();
                    let mutex = Arc::new(Mutex::new((VecDeque::new(), send.clone(), recv, target.clone())));
                    let task = spawn(async move {
                        let mut buf = [0u8; BUF];
                        loop {
                            match target.read(&mut buf).await {
                                Ok(0) | Err(_) => {
                                    let _ = send.send(Ok(true)).await;
                                    return;
                                }
                                Ok(n) => if send.send(Err(buf[0..n].iter().copied().collect())).await.is_err() { return; },
                            }
                        }
                    });
                    connections.lock().await.insert(connection_id, (task, Arc::clone(&mutex)));
                    mutex
                }
                Err(e) => {
                    eprintln!("[{connection_id}] Couldn't connect to {name} ({addr}): {e}");
                    return Ok(());
                }
            }
        } else {
            eprintln!("[{connection_id}] No such name: {name}");
            return Ok(());
        };
        eprintln!("[{connection_id}] connected");
        o
    } else {
        let mut b = [0u8; 8];
        connection.read_exact(&mut b).await?;
        connection_id = u64::from_be_bytes(b);
        // reconnection
        let o = if let Some((_task, mutex)) = connections.lock().await.get(&connection_id) {
            Arc::clone(mutex)
        } else {
            eprintln!("[{connection_id}] no connection with the requested id exists");
            return Ok(());
        };
        eprintln!("[{connection_id}] reconnected");
        o
    };
    let mut lock = if let Some(lock) = mutex.try_lock() {
        lock
    } else {
        eprintln!("[{connection_id}] waiting for lock...");
        let lock = mutex.lock().await;
        lock
    };
    let _task_fwd = {
        let mut remote = connection.clone();
        let mut local = lock.3.clone();
        let send = lock.1.clone();
        spawn(async move {
            let mut buf = [0u8; BUF];
            loop {
                match remote.read(&mut buf).await {
                    Ok(0) => {
                        // EOF -> Disconnect
                        let _ = send.send(Ok(false)).await;
                        return;
                    }
                    Ok(n) => if local.write_all(&buf[0..n]).await.is_err() {
                        // local disconnect
                        let _ = send.send(Ok(true)).await;
                        return;
                    },
                    Err(_) => {
                        // reconnect
                        let _ = send.send(Ok(false)).await;
                        return;
                    }
                };
            }
        })
    };
    'fwd: while let Ok(data) = if lock.0.is_empty() { lock.2.recv().await } else { Ok(Err(std::mem::replace(&mut lock.0, VecDeque::new()))) } {
        if let Err(mut data) = data {
            while !data.is_empty() {
                match connection.write(data.make_contiguous()).await {
                    Ok(0) => break 'fwd,
                    Ok(n) => for _ in 0..n {
                        data.pop_front();
                    }
                    Err(_) => {
                        lock.0 = data;
                        eprintln!("[{connection_id}] transport connection closed temporarily");
                        return Ok(());
                    }
                }
            }
        } else if let Ok(true) = data {
            break;
        } else {
            eprintln!("[{connection_id}] transport connection closed temporarily");
            return Ok(());
        }
    }
    eprintln!("[{connection_id}] connection closed and removed");
    connections.lock().await.remove(&connection_id);
    Ok(())
}

pub async fn client(local: TcpStream, remote: SocketAddr, name: String, kill: Option<Arc<AtomicBool>>) -> io::Result<()> {
    if name.len() > 254 {
        eprintln!("[ERR] name too long");
        Err(io::ErrorKind::InvalidInput)?;
    }
    let mut connection_id: Option<u64> = None;
    let mut send_buf = VecDeque::new();
    'reconnect: loop {
        let mut remote = loop {
            if !kill.as_ref().is_some_and(|kill| kill.load(Ordering::Relaxed)) {
                match TcpStream::connect(remote).await {
                    Ok(stream) => break stream,
                    Err(e) => eprintln!("[{}] Could not connect to {remote}: {e}", connection_id.map(|v| v.to_string()).as_deref().unwrap_or("-")),
                }
            }
            Timer::after(Duration::from_secs(3)).await;
        };
        let mut init_msg = [0u8; 16];
        remote.read_exact(&mut init_msg).await?;
        if init_msg[0..5] != [b't', b'm', b'o', b't', b'h'] {
            eprintln!("This does not appear to be a tmoth socket (got {})...", String::from_utf8_lossy(&init_msg[0..5]));
            Err(io::ErrorKind::InvalidData)?;
        }
        let major = init_msg[5];
        let minor = init_msg[6];
        let patch = init_msg[7];
        if MAJOR > major {
            eprintln!("[VERSION] Server has older MAJOR version: {major} vs. {MAJOR}");
            Err(io::ErrorKind::Unsupported)?;
        } else if MAJOR < major {
            eprintln!("[VERSION] Server has newer MAJOR version: {major} vs. {MAJOR}");
            Err(io::ErrorKind::Unsupported)?;
        }
        if MINOR > minor {
            eprintln!("[version] Server has older minor version: {minor} vs. {MINOR}");
        } else if MINOR < minor {
            eprintln!("[version] Server has newer minor version: {minor} vs. {MINOR}");
        }
        if PATCH > patch {
            eprintln!("[version] Server has older patch version: {patch} vs. {PATCH}");
        } else if PATCH < patch {
            eprintln!("[version] Server has newer patch version: {patch} vs. {PATCH}");
        }
        let connection_id = if let Some(connection_id) = connection_id {
            // reconnect
            let bytes: [u8; 8] = connection_id.to_be_bytes();
            if remote.write_all(&[0xFF, bytes[0], bytes[1], bytes[2], bytes[3], bytes[4], bytes[5], bytes[6], bytes[7]]).await.is_err() {
                Timer::after(Duration::from_secs(3)).await;
                continue 'reconnect;
            }
            eprintln!("[{connection_id}] reconnected");
            connection_id
        } else {
            // initial connect
            let cid = u64::from_be_bytes([init_msg[8], init_msg[9], init_msg[10], init_msg[11], init_msg[12], init_msg[13], init_msg[14], init_msg[15]]);
            remote.write_all(&[name.len() as u8]).await?;
            remote.write_all(name.as_bytes()).await?;
            eprintln!("[{cid}] connected");
            connection_id = Some(cid);
            cid
        };
        let (send, recv) = unbounded();
        let _task = {
            let mut local = local.clone();
            let send = send.clone();
            spawn(async move {
                let mut buf = [0u8; BUF];
                loop {
                    match local.read(&mut buf).await {
                        Ok(0) | Err(_) => {
                            let _ = send.send(Ok(true)).await;
                            return;
                        }
                        Ok(n) => if send.send(Err(buf[0..n].iter().copied().collect())).await.is_err() { return; },
                    }
                }
            })
        };
        let _task_fwd = {
            let mut remote = remote.clone();
            let mut local = local.clone();
            let kill = kill.as_ref().map(Arc::clone);
            spawn(async move {
                let mut buf = [0u8; BUF];
                loop {
                    if kill.as_ref().is_some_and(|kill| kill.load(Ordering::Relaxed)) {
                        let _ = send.send(Ok(false)).await;
                        return;
                    } else {
                        match remote.read(&mut buf).await {
                            Ok(0) => {
                                // EOF -> Disconnect
                                let _ = send.send(Ok(false)).await;
                                return;
                            }
                            Ok(n) => if local.write_all(&buf[0..n]).await.is_err() {
                                // local disconnect
                                let _ = send.send(Ok(true)).await;
                                return;
                            },
                            Err(_) => {
                                // reconnect
                                let _ = send.send(Ok(false)).await;
                                return;
                            }
                        }
                    }
                }
            })
        };
        'fwd: while let Ok(data) = if send_buf.is_empty() { recv.recv().await } else { Ok(Err(std::mem::replace(&mut send_buf, VecDeque::new()))) } {
            if let Err(mut data) = data {
                let kill = kill.as_ref().is_some_and(|kill| kill.load(Ordering::Relaxed));
                while !data.is_empty() {
                    match if kill { None } else { remote.write(data.make_contiguous()).await.ok() } {
                        Some(0) => break 'fwd,
                        Some(n) => for _ in 0..n {
                            data.pop_front();
                        }
                        None => {
                            send_buf = data;
                            eprintln!("[{connection_id}] disconnected temporarily");
                            Timer::after(Duration::from_secs(3)).await;
                            continue 'reconnect;
                        }
                    }
                }
            } else if let Ok(true) = data {
                return Ok(());
            } else {
                eprintln!("[{connection_id}] temporary disconnect");
                Timer::after(Duration::from_secs(3)).await;
                continue 'reconnect;
            }
        }
    }
}
